<?php

/**
 * @file
 * Provides Drupal implementation of project release info parsing and batch API.
 */

/**
 * Regular expression to match PHP function names. (From Drupal 7)
 *
 * @see http://php.net/manual/en/language.functions.php
 */
define('DRUPAL_PHP_FUNCTION_PATTERN', '[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*');

module_load_include('package.inc', 'project_dependency');

/**
 * Minimum API compatibility to process. We will not look at modules before
 * Drupal 6.
 */
define('PROJECT_DEPENDENCY_MINIMUM_DRUPAL_VERSION', 6);

/**
 * Directory for git repositories to be processed.
 *
 * These get checked out automatically, but the whole process goes faster
 * if it's already prepopulated with everything we need.
 *
 */
define('PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY', variable_get('project_dependency_sourcecode_directory', '/tmp/project_dependency_sourcecode_directory'));

/**
 * Path to the upstream git repository.
 *
 * If the source code repository of any project is not found locally, it will be
 * fetched from this path.
 */
define('PROJECT_DEPENDENCY_UPSTREAM_REPOSITORY_PATH', variable_get('project_dependency_upstream_repository_path', 'git://git.drupal.org/project'));

/**
 * Parse all the .info files.
 *
 * @param $info_files
 *   List of component .info files.
 *
 * @return
 *   Associative array of component information keyed by component name and
 *   merged with an array of defaults.
 */
function project_dependency_info_parse(array $info_files) {
  $defaults = array(
    'name' => 'Unknown',
    'description' => '',
    'dependencies' => array(),
    'test_dependencies' => array(),
  );

  $info = array();
  foreach ($info_files as $file) {
    $component = basename($file, '.info');
    $info[$component] = drupal_parse_info_file($file) + $defaults;

    // Change info keys to suite project_dependency_info.
    $info[$component]['title'] = $info[$component]['name'];
    $info[$component]['name'] = $component;
  }
  return $info;
}

/**
 * Get the project release nodes that contain a component and are compatible
 * with an API, ordered by most recent (highest nid)
 *
 * Sadly, the component may appear in more than one product, since there's no
 * guaranteed uniqueness of a submodule name. So we're just preferring the
 * most recent release node (highest nid).
 *
 * @param $component
 *   Component name.
 * @param $api_tid
 *   Core API compatibility tid.
 *
 * @return
 *   Array of release arrays, keyed by project release node nid, each of which
 *   has the keys
 *   - 'uri'           Project URI (shortname)
 *   - 'component_id'  Numeric key
 *   - 'pid'           Project nid
 *   - 'release_nid'
 *   - 'version'       Release name
 *   - 'tag'           VCS tag
 *   - 'version_major' project_release_node.version_major
 *   . Each indicates that the given component
 *   name is found in the given release. These are ordered by version string
 *   descending.
 */
function project_dependency_info_releases_for_component_get($component, $api_tid) {
  $result = db_query("SELECT DISTINCT pp.uri, pp.nid AS pid, c.component_id, c.release_nid, prn.version, prn.tag, prn.version_major
    FROM {project_dependency_component} c
    INNER JOIN {project_release_nodes} prn ON c.release_nid = prn.nid
    LEFT JOIN {project_projects} pp ON prn.pid = pp.nid
    INNER JOIN {term_node} t ON c.release_nid = t.nid
    WHERE t.tid = %d AND c.name = '%s'
    ORDER BY prn.nid DESC, t.vid DESC
  ", $api_tid, $component);

  $releases = array();
  while ($release = db_fetch_array($result)) {
    $releases[$release['release_nid']] = $release;
  }
  return $releases;
}

/**
 * Attempt to determine the Drupal core API term.
 *
 * @param $node
 *   Node object.
 *
 * @return
 *   Core API term to which the node belongs, otherwise FALSE.
 */
function project_dependency_info_core_api($node) {
  static $api_terms = array();

  if (empty($node->nid)) {
    return NULL;
  }
  if (empty($api_terms[$node->nid])) {
    $api_terms[$node->nid] = FALSE;

    if (!empty($node->taxonomy)) {
      // Relase API version vocabulary.
      $api_vid = _project_release_get_api_vid();

      foreach ($node->taxonomy as $tid => $term) {
        if ($term->vid == $api_vid) {
          $api_terms[$node->nid] = $term;
        }
      }
    }
  }
  return $api_terms[$node->nid];
}

/**
 * Given a core API taxonomy term name like '6.x', return the numeric core api
 * associated with it.
 *
 * @param $core_api_term_name
 */
function project_dependency_core_api_from_term($core_api_term_name) {
  return preg_replace('/\.x.*$/', '', $core_api_term_name);
}

/**
 * Check out a release and determine relevant information.
 *
 * Locate all .info files contained within the project, determine the names of
 * the module(s) contained within the project and store the dependencies for
 * later processing.
 *
 * @param $release
 *   A keyed array of release information, containing
 *   - 'tag' => The tag of the release (like 7.x-1.0)
 *   - 'uri' => The project shortname (like 'views')
 *   - 'nid' => The nid of the project release node
 *
 * @return
 *   An array of dependencies, or FALSE on failure.
 */
function project_dependency_info_batch_process_release(array $release) {
  // Use batch directory.
  if (!is_dir(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY)) {
    if (!mkdir(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY)) {
      watchdog('project_dependency', 'Failed to create PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY %directory', array('%directory' => PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY));
      return FALSE;
    }
  }

  // Checkout release.
  $url = PROJECT_DEPENDENCY_UPSTREAM_REPOSITORY_PATH . '/' . escapeshellcmd($release['uri']) . '.git';
  $tag = escapeshellcmd($release['tag']);
  $directory = $release['uri'];

  $work_tree = PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY . "/$directory";
  $git_dir = "$work_tree/.git";
  $esc_git_dir = escapeshellarg($git_dir);
  $work_tree = escapeshellarg($work_tree);
  // If the config file in git dir doesn't exist, then remove the whole thing.
  if (!is_file("$git_dir/config")) {
    $command = "rm -rf $work_tree 2>&1";
    exec($command, $output, $status);
    if ($status) {
      watchdog('project_dependency', 'Failed to execute command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
      return FALSE;
    }
    // Clone the repo
    $command = "git clone $url $work_tree 2>&1";
    exec($command, $output, $status);
    if ($status) {
      watchdog('project_dependency', 'Failed to execute git command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
      return FALSE;
    }
  }
  // But if git dir already existed and looked healthy
  else {
    $command = "git --work-tree $work_tree --git-dir $esc_git_dir fetch 2>&1";
    exec($command, $output, $status);
    if ($status) {
      watchdog('project_dependency', 'Failed to execute git command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
      return FALSE;
    }
  }

  // We need to make sure the directory is clean, so reset and clean
  // git reset --hard
  exec("git --work-tree $work_tree --git-dir $esc_git_dir reset --hard 2>&1", $output, $status);
  // git clean -fd
  exec("git --work-tree $work_tree --git-dir $esc_git_dir clean -fd 2>&1", $output, $status);
  // Make sure we don't have an existing branch by this name, or it won't be updated properly.
  // First get into a detached-head state so we know we can delete the branch.
  unset($output);
  exec("git --work-tree=$work_tree --git-dir $esc_git_dir rev-list --max-count=1 HEAD", $output, $status);
  $hash = $output[0];
  exec("git --work-tree=$work_tree --git-dir $esc_git_dir checkout '$hash' 2>&1", $output, $status);

  // Then authoritatively delete any branch by that name. OK for this to fail.
  unset($output);
  exec("git --work-tree=$work_tree --git-dir $esc_git_dir branch -D '$tag' 2>&1", $output, $status);

  // git checkout tag
  $command = "git --work-tree=$work_tree --git-dir $esc_git_dir checkout '$tag' 2>&1";
  unset($output);
  exec($command, $output, $status);
  if ($status) {
    watchdog('project_dependency', 'Failed to execute git command %cmd; output=%output', array('%cmd' => $command, '%output' => implode(' ', $output)));
    return FALSE;
  }

  // Scan checkout for .info files and create a list in the same format that
  // project release uses, so that standard API functions can be used.

  // Note that DRUPAL_PHP_FUNCTION_PATTERN croaks in file_scan_directory()
  // "WD php: Warning: ereg(): REG_ERANGE in file_scan_directory()"
  // $function_pattern = DRUPAL_PHP_FUNCTION_PATTERN;
  $function_pattern = '[a-zA-Z][a-zA-Z0-9_]*';
  $profile_info_files = file_scan_directory(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY . '/' . $directory, '^' . $function_pattern . '\.(theme|profile)$');
  // If any profile or theme files are found, skip this release entirely.
  if (!empty($profile_info_files)) {
    return FALSE;
  }
  $files = file_scan_directory(PROJECT_DEPENDENCY_SOURCECODE_DIRECTORY . '/' . $directory, '^' . $function_pattern . '\.info$');
  $info_files = array();
  foreach ($files as $file) {
    $info_files[] = $file->filename;
  }
  $info = project_dependency_info_parse($info_files);

  $dependencies = array();
  foreach ($info as $module => $module_info) {
    $dependencies[$module] = $module_info['dependencies'];
    $test_dependencies[$module] = $module_info['test_dependencies'];
  }

  // Clear previous records for the release.
  project_dependency_info_package_clear($release['nid']);

  $components_in_release = array_keys($info);
  // Store the list of components contained by the project.
  $component_info = project_dependency_info_package_list_store($release['nid'], $info);
  foreach ($component_info as $component => $item) {
    // First process real, required dependencies
    $new_dependencies = array();
    if (is_array($info[$component]['dependencies'])) {
      foreach ($info[$component]['dependencies'] as $new_dependency) {
        $new_dependencies[$new_dependency] = array('external' => !in_array($new_dependency, $components_in_release));
      }
    }
    project_dependency_info_package_dependencies_store($item['release_nid'], $item['component_id'], $new_dependencies);

    // Now process test_dependencies as "RECOMMENDED"
    $test_dependencies = array();
    if (is_array($info[$component]['test_dependencies'])) {
      foreach ($info[$component]['test_dependencies'] as $new_dependency) {
        $test_dependencies[$new_dependency] = array('external' => !in_array($new_dependency, $components_in_release));
      }
    }
    project_dependency_info_package_dependencies_store($item['release_nid'], $item['component_id'], $test_dependencies, PROJECT_DEPENDENCY_DEPENDENCY_RECOMMENDED);
  }

  return $dependencies;
}


/**
 * Build dependency information for a single release node if it meets the
 * minimum Drupal major version API requirement.
 *
 * @param $shortname
 *   The shortname of the project.
 * @param $version
 *   The version of the project
 *
 * @return
 *   An array of dependency information or FALSE on failure.
 */
function project_dependency_process_release($shortname, $version) {
  // Load the project.
  $pid = project_get_nid_from_uri($shortname);
  $query = 'SELECT nid,tag FROM {project_release_nodes} WHERE pid = %d AND version = "%s"';
  $release = db_fetch_array(db_query($query, $pid, $version));
  $release['uri'] = $shortname;

  $node = node_load($release['nid']);
  if (empty($node)) {
    return FALSE;
  }
  $core_api_term = project_dependency_info_core_api($node);
  if (empty($core_api_term)) {
    return FALSE;
  }

  $core_api = project_dependency_core_api_from_term($core_api_term->name);
  if ($core_api < PROJECT_DEPENDENCY_MINIMUM_DRUPAL_VERSION) {
    return FALSE;
  }

  $dependencies = project_dependency_info_batch_process_release($release);
  return $dependencies;
}

/**
 * Find all dependencies (as release nids) of a given release nid and optional
 * components.
 *
 * If $depending_components are provided, only their dependencies will be
 * returned, rather than the dependencies of every component in the release.
 *
 * OK, I'm sorry this is a mindbending function. I've tried to compensate with
 * lots of comments, but suspect it's nowhere near adequate.
 *
 * This is a recursive, accumulative function, gathering the results we need in
 * $dependency_releases.
 *
 * @param $release_nid
 *   nid of the release node.
 * @param $depending_components
 *   an array of component names, like
 *   array('views_ui', 'token', 'page_manager')
 * @param $dependency_releases (by reference)
 *   The array of release nids that are depended upon by the given components.
 * @param $recursion_level
 *   Recursion level of this call to prevent infinite recursion on some
 *   database problem.
 *
 * Same as  project_dependency_info_process_dependencies($rid, $modules[$module], $dependencies);
 */
function project_dependency_get_external_component_dependencies($release_nid, $depending_components, &$dependency_releases, $recursion_level = 0) {
  // Limit this to 4 levels for now, as we probably don't know of anything
  // that needs recursion_level > 2.
  if ($recursion_level > 4) {
    return;
  }
  // Get the release node and project node of the *depending* components.
  $release_node = node_load($release_nid);
  $project_node = node_load($release_node->project_release['pid']);

  // We don't need to study the dependencies of the drupal project.
  if ($project_node->project['uri'] == 'drupal') {
    return;
  }

  // Get the api_term of the *depending* components.
  if (!($api_term = project_dependency_info_core_api($release_node))) {
    watchdog('project_dependency', 'ERROR: No core release API term found for release nid=%nid', array('%nid' => $release_nid));
    return;
  }

  // If no list of depending components was passed in, create a list of all
  // components in this release.
  if (empty($depending_components)) {
    $query = 'SELECT name FROM {project_dependency_component} WHERE release_nid = %d';
    $result = db_query($query, $release_nid);
    while ($row = db_fetch_array($result)) {
      $depending_components[] = $row['name'];
    }
  }
  // Now we have a list of component names for which we need to find
  // external dependencies. Walk through them and for each one, find the
  // external (not in this release) dependencies.
  foreach ($depending_components as $depending_component) {
    // Get a list of keyed arrays the describe external dependencies.
    $external_dependencies = project_dependency_find_external_dependencies($release_nid, $depending_component);

    if (empty($external_dependencies)) {
      continue;
    }

    // Walk through the external dependencies to find releases they are shipped
    // in.
    foreach ($external_dependencies as $component_info) {
      $dependency_component = $component_info['dependency'];
      $parsed_component =
        project_dependency_parse_dependency($dependency_component, $api_term->name);
      $dependency_component = $parsed_component['name'];

      // Find the releases that component was shipped in.
      // We could hope that it would only be in one package/rid,
      // but there are probably cases (refactoring) where that is not
      // the case.
      $possibilities = project_dependency_info_releases_for_component_get($dependency_component, $api_term->tid);
      // In the case we found no possible release, we must halt processing and
      // continue to the next item
      if (empty($possibilities)) {
        watchdog('project_dependency', 'Failed to find a release for component %component as dependency of %depending_component, release_nid=%release_nid', array('%component' => $dependency_component, '%depending_component' => $depending_component, '%release_nid' => $release_nid));
        continue;
      }

      $possible_dependent_releases = array();
      foreach ($possibilities as $dependent_release) {
        // Remove the api-term from the front of the version if necessary.
        // The complex regex is for the case that we have a 7.x-7.x branch
        // specified specified as just (7.x). We don't want it removed in that
        // case.
        $version_to_check = preg_replace("/^{$api_term->name}-([0-9]+\.[0-9x]+.*$)/", '$1', $dependent_release['version']);
        // Remove -dev from the end of a dependency like (1.x-dev)
        $version_to_check = preg_replace('/\.x-dev$/', '.x', $version_to_check);
        $check = project_dependency_check_incompatibility($parsed_component, $version_to_check);
        if (is_null($check)) {
          $possible_dependent_releases[$dependent_release['release_nid']] =
            $dependent_release;
        }
      }

      // We choose to use the project of the most recently created release
      // node. This is not guaranteed, but is a heuristic. Is there another?
      $first_dependent_release = reset($possible_dependent_releases);
      $dependent_pid = $first_dependent_release['pid'];
      $recommended_release = project_release_get_current_recommended($dependent_pid, $api_term->tid);

      // If we found a recommended release for this project and that release
      // also contains this component, accept it.
      if (!empty($recommended_release) && !empty($possible_dependent_releases[$recommended_release->nid])) {
        // Accept the recommended release for the project used.
        $release = $possible_dependent_releases[$recommended_release->nid];
      }
      else {
        // The recommended release does not contain the component, so just
        // punt and accept the first release that was found.
        $release = $first_dependent_release;
      }
      $release['component'] = $dependency_component;
      $release['dependency_type'] = $component_info['dependency_type'];

      // If the release has already been listed and is REQUIRED, we don't want
      // to overwrite it, so will not overwrite with a RECOMMENDED release.
      // Also will give first precedence to the first release found for component.
      if (empty($dependency_releases[$dependency_component]) || $dependency_releases[$dependency_component]['dependency_type'] == PROJECT_DEPENDENCY_DEPENDENCY_RECOMMENDED) {
        $dependency_releases[$dependency_component] = $release;
      }
      // Recurse to find dependencies of the discovered dependency, but only
      // for the component which discovered the dependent release.
      project_dependency_get_external_component_dependencies($release['release_nid'], array($dependency_component), $dependency_releases, $recursion_level + 1);
    }
  }
}


/**
 * Given a release nid and component name, find all external dependencies for
 * that component.
 * @param $release_nid
 *   Release nid.
 * @param $component_name
 *   Component name.
 *
 * @return
 *   An array of component information.
 */
function project_dependency_find_external_dependencies($release_nid, $component_name) {
  $rows = array();
  $query = 'SELECT DISTINCT(pdc.component_id), name AS depender, dependency, release_nid, dependency_type FROM {project_dependency_dependency} pdd INNER JOIN {project_dependency_component} pdc
            ON pdd.component_id = pdc.component_id WHERE  pdc.name = "%s" AND pdc.release_nid = %d';
  $result = db_query($query, $component_name, $release_nid);
  while($row = db_fetch_array($result)) {
    $rows[] = $row;
  }
  return $rows;
}
